# Tutorial setup

import argparse
import os
import time
import numpy as np
import gzip
import requests
import json
import tensorflow as tf


class MNIST:
    def extract_data(self, filename, num_images):
        with gzip.open(filename) as bytestream:
            bytestream.read(16)
            buf = bytestream.read(num_images * 28 * 28)
            data = np.frombuffer(buf, dtype=np.uint8).astype(np.float32)
            data = (data / 255) - 0.5
            data = data.reshape(num_images, 28, 28, 1)
            return data

    def extract_labels(self, filename, num_images):
        with gzip.open(filename) as bytestream:
            bytestream.read(8)
            buf = bytestream.read(1 * num_images)
            labels = np.frombuffer(buf, dtype=np.uint8)
        return labels.astype(np.uint8)

    def __init__(self):
        data_path = "data/FashionMNIST/raw"
        self.train_size = 60000
        self.test_size = 10000
        self.train_data = {
            "x": self.extract_data(
                os.path.join(data_path, "train-images-idx3-ubyte.gz"), self.train_size
            ),
            "y": self.extract_labels(
                os.path.join(data_path, "train-labels-idx1-ubyte.gz"), self.train_size
            ),
        }
        self.test_data = {
            "x": self.extract_data(
                os.path.join(data_path, "t10k-images-idx3-ubyte.gz"), self.test_size
            ),
            "y": self.extract_labels(
                os.path.join(data_path, "t10k-labels-idx1-ubyte.gz"), self.test_size
            ),
        }


def chief_completed():
    k8s_sa_path = "/var/run/secrets/kubernetes.io/serviceaccount"
    with open(os.path.join(k8s_sa_path, "namespace"), "r") as f:
        namespace = f.read()
    with open(os.path.join(k8s_sa_path, "token"), "r") as f:
        token = f.read()
    headers = {"Authorization": "Bearer {}".format(token)}
    chief_pod_name = json.loads(os.environ["TF_CONFIG"])["cluster"]["chief"][0].split(
        ":"
    )[0]
    response = requests.request(
        "GET",
        "https://kubernetes.default.svc/api/v1/namespaces/{}/pods/{}".format(
            namespace, chief_pod_name
        ),
        headers=headers,
        verify=os.path.join(k8s_sa_path, "ca.crt"),
    )
    raw_predictions = json.loads(response.text)
    return raw_predictions["status"]["phase"] == "Succeeded"


def main():
    # Training settings
    parser = argparse.ArgumentParser(description="Tensorflow MNIST Example")
    parser.add_argument(
        "--batch-size",
        type=int,
        default=64,
        metavar="N",
        help="input batch size for training (default: 64)",
    )
    parser.add_argument(
        "--test-batch-size",
        type=int,
        default=1000,
        metavar="N",
        help="input batch size for testing (default: 1000)",
    )
    parser.add_argument(
        "--epochs",
        type=int,
        default=10,
        metavar="N",
        help="number of epochs to train (default: 10)",
    )
    parser.add_argument(
        "--lr",
        type=float,
        default=0.01,
        metavar="LR",
        help="learning rate (default: 0.01)",
    )
    parser.add_argument(
        "--momentum",
        type=float,
        default=0.5,
        metavar="M",
        help="SGD momentum (default: 0.5)",
    )
    args = parser.parse_args()

    # Set the environment variable to allow reporting worker and ps failure to the
    # coordinator. This is a workaround and won't be necessary in the future.
    os.environ["GRPC_FAIL_FAST"] = "use_caller"

    cluster_resolver = tf.distribute.cluster_resolver.TFConfigClusterResolver()

    if cluster_resolver.task_type in ("worker", "ps"):
        # Start a TensorFlow server and wait.
        server = tf.distribute.Server(
            cluster_resolver.cluster_spec(),
            job_name=cluster_resolver.task_type,
            task_index=cluster_resolver.task_id,
            protocol=cluster_resolver.rpc_layer or "grpc",
            start=True,
        )

        while not chief_completed():
            os.system('nvidia-smi')
            time.sleep(10)
        exit()

    else:
        # Run the coordinator.
        # Dispatch training steps to remote workers

        # Instantiate a ParameterServerStrategy

        variable_partitioner = (
            tf.distribute.experimental.partitioners.MinSizePartitioner(
                min_shard_bytes=(256 << 10),
                max_shards=len(json.loads(os.environ["TF_CONFIG"])["cluster"]["ps"]),
            )
        )

        strategy = tf.distribute.ParameterServerStrategy(
            cluster_resolver, variable_partitioner=variable_partitioner
        )

        # Set up the data

        mnist = MNIST()

        dataset = (
            tf.data.Dataset.from_tensor_slices(
                (mnist.train_data["x"], mnist.train_data["y"])
            )
            .shuffle(mnist.train_size)
            .batch(args.batch_size)
            .repeat()
            .prefetch(2)
        )

        # Build the model
        with strategy.scope():
            # Create the model. The input needs to be compatible with Keras processing layers.
            model = tf.keras.models.Sequential(
                [
                    tf.keras.layers.InputLayer(input_shape=(28, 28, 1)),
                    tf.keras.layers.Conv2D(
                        20, (5, 5), activation="relu", padding="valid", strides=1
                    ),
                    tf.keras.layers.MaxPooling2D((2, 2)),
                    tf.keras.layers.Conv2D(
                        50, (5, 5), activation="relu", padding="valid", strides=1
                    ),
                    tf.keras.layers.MaxPooling2D((2, 2)),
                    tf.keras.layers.Flatten(),
                    tf.keras.layers.Dense(500, activation="relu"),
                    tf.keras.layers.Dense(10),
                ]
            )

            model.compile(
                optimizer=tf.keras.optimizers.SGD(
                    learning_rate=args.lr, momentum=args.momentum
                ),
                loss=tf.keras.losses.SparseCategoricalCrossentropy(from_logits=True),
                metrics=[tf.keras.metrics.SparseCategoricalAccuracy()],
            )

        model.fit(
            dataset,
            epochs=args.epochs,
            steps_per_epoch=int(mnist.train_size / args.batch_size),
        )

        coordinator = tf.distribute.coordinator.ClusterCoordinator(strategy)

        # Inline evaluation

        with strategy.scope():
            # Define the eval metric on parameter servers.
            eval_accuracy = tf.keras.metrics.SparseCategoricalAccuracy()

        @tf.function
        def eval_step(iterator):
            def replica_fn(batch_data, labels):
                pred = model(batch_data, training=False)
                eval_accuracy.update_state(labels, pred)

            batch_data, labels = next(iterator)
            strategy.run(replica_fn, args=(batch_data, labels))

        def eval_dataset_fn():
            return (
                tf.data.Dataset.from_tensor_slices(
                    (mnist.test_data["x"], mnist.test_data["y"])
                )
                .shuffle(mnist.test_size)
                .batch(args.test_batch_size)
                .repeat()
            )

        per_worker_eval_dataset = coordinator.create_per_worker_dataset(eval_dataset_fn)
        per_worker_eval_iterator = iter(per_worker_eval_dataset)

        eval_steps_per_epoch = int(mnist.test_size / args.test_batch_size)
        for _ in range(eval_steps_per_epoch):
            coordinator.schedule(eval_step, args=(per_worker_eval_iterator,))
        coordinator.join()
        print("Evaluation accuracy: %f" % eval_accuracy.result())


if __name__ == "__main__":
    main()
